@@ -45,11 +45,11 @@ static FrCommandClass *parent_class = NULL;
 /* -- list -- */
 
 static time_t
-mktime_from_string (char *month, 
+mktime_from_string (char *month,
 		    char *mday,
 		    char *year)
 {
-	static char  *months[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", 
+	static char  *months[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
 				   "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
 	struct tm     tm = {0, };
 
@@ -72,7 +72,7 @@ mktime_from_string (char *month,
 
 	  		tm.tm_hour = atoi (fields[0]);
 	  		tm.tm_min = atoi (fields[1]);
-	  
+
 	  		now = time(NULL);
 	  		now_tm = localtime (&now);
 	  		tm.tm_year = now_tm->tm_year;
@@ -85,19 +85,18 @@ mktime_from_string (char *month,
 
 
 static void
-list__process_line (char     *line, 
+list__process_line (char     *line,
 		    gpointer  data)
 {
 	FileData    *fdata;
 	FrCommand   *comm = FR_COMMAND (data);
 	char       **fields;
 	const char  *name_field;
+	char        *name;
+	int          ofs = 0;
 
 	g_return_if_fail (line != NULL);
 
-	if (line[0] == 'd') /* Ignore directories. */
-		return;
-
 	fdata = file_data_new ();
 
 #ifdef __sun
@@ -107,14 +106,23 @@ list__process_line (char     *line,
 	g_strfreev (fields);
 
 	name_field = get_last_field (line, 10);
-#else
-	fields = split_line (line, 8);
-	fdata->size = g_ascii_strtoull (fields[4], NULL, 10);
-	fdata->modified = mktime_from_string (fields[5], fields[6], fields[7]);
+#else /* !__sun */
+	/* Handle char and block device files */
+	if ((line[0] == 'c') || (line[0] == 'b')) {
+		fields = split_line (line, 9);
+		ofs = 1;
+		fdata->size = 0;
+		/* TODO We should also specify the content type */
+	}
+	else {
+		fields = split_line (line, 8);
+		fdata->size = g_ascii_strtoull (fields[4], NULL, 10);
+	}
+	fdata->modified = mktime_from_string (fields[5+ofs], fields[6+ofs], fields[7+ofs]);
 	g_strfreev (fields);
 
-	name_field = get_last_field (line, 9);
-#endif /* __sun */
+	name_field = get_last_field (line, 9+ofs);
+#endif /* !__sun */
 
 	fields = g_strsplit (name_field, " -> ", 2);
 
@@ -123,45 +131,51 @@ list__process_line (char     *line,
 		fields = g_strsplit (name_field, " link to ", 2);
 	}
 
+	fdata->dir = line[0] == 'd';
+
+	name = g_strcompress (fields[0]);
 	if (*(fields[0]) == '/') {
-		fdata->full_path = g_strcompress (fields[0]);
+		fdata->full_path = g_strdup (name);
 		fdata->original_path = fdata->full_path;
-	} 
+	}
 	else {
-		char *compressed;
-		
-		compressed = g_strcompress (fields[0]);
-		fdata->full_path = g_strconcat ("/", compressed, NULL);
-		g_free (compressed);
-		
+		fdata->full_path = g_strconcat ("/", name, NULL);
 		fdata->original_path = fdata->full_path + 1;
 	}
+	if (fdata->dir && (name[strlen (name) - 1] != '/')) {
+		char *old_full_path = fdata->full_path;
+		fdata->full_path = g_strconcat (old_full_path, "/", NULL);
+		g_free (old_full_path);
+		fdata->original_path = g_strdup (name);
+		fdata->free_original_path = TRUE;
+	}
+	g_free (name);
 
 	if (fields[1] != NULL)
 		fdata->link = g_strcompress (fields[1]);
 	g_strfreev (fields);
 
-	fdata->name = g_strdup (file_name_from_path (fdata->full_path));
+	if (fdata->dir)
+		fdata->name = dir_name_from_path (fdata->full_path);
+	else
+		fdata->name = g_strdup (file_name_from_path (fdata->full_path));
 	fdata->path = remove_level_from_path (fdata->full_path);
 
 	if (*fdata->name == 0)
 		file_data_free (fdata);
-	else 
+	else
 		fr_command_add_file (comm, fdata);
 }
 
 
 static void
-fr_command_rpm_list (FrCommand  *comm,
-		     const char *password)
+fr_command_rpm_list (FrCommand *comm)
 {
-	fr_process_set_out_line_func (FR_COMMAND (comm)->process, 
-				      list__process_line,
-				      comm);
+	fr_process_set_out_line_func (comm->process, list__process_line, comm);
 
-	fr_process_begin_command (comm->process, "rpm2cpio");
-	fr_process_add_arg (comm->process, comm->e_filename);
-	fr_process_add_arg (comm->process, "| cpio -itv");
+	fr_process_begin_command (comm->process, "sh");
+	fr_process_add_arg (comm->process, "-c");
+	fr_process_add_arg_concat (comm->process, PRIVEXECDIR "rpm2cpio ", comm->e_filename, " -itv", NULL);
 	fr_process_end_command (comm->process);
 	fr_process_start (comm->process);
 }
@@ -169,42 +183,72 @@ fr_command_rpm_list (FrCommand  *comm,
 
 static void
 fr_command_rpm_extract (FrCommand  *comm,
+		        const char  *from_file,
 			GList      *file_list,
 			const char *dest_dir,
 			gboolean    overwrite,
 			gboolean    skip_older,
-			gboolean    junk_paths,
-			const char *password)
+			gboolean    junk_paths)
 {
-	GList *scan;
+	GList   *scan;
+	GString *cmd;
 
-	fr_process_begin_command (comm->process, "rpm2cpio");
+	fr_process_begin_command (comm->process, "sh");
 	if (dest_dir != NULL)
                 fr_process_set_working_dir (comm->process, dest_dir);
-	fr_process_add_arg (comm->process, comm->e_filename);
-	fr_process_add_arg (comm->process, "| cpio -idu");
+	fr_process_add_arg (comm->process, "-c");
+
+	cmd = g_string_new (PRIVEXECDIR "rpm2cpio ");
+	g_string_append (cmd, comm->e_filename);
+	g_string_append (cmd, " -idu ");
 	for (scan = file_list; scan; scan = scan->next) {
-		char *filename = (char*) scan->data;
-		fr_process_add_arg (comm->process, filename);
+		char *filename = g_shell_quote (scan->data);
+		g_string_append (cmd, filename);
+		g_free (filename);
+		g_string_append (cmd, " ");
 	}
+	fr_process_add_arg (comm->process, cmd->str);
+	g_string_free (cmd, TRUE);
+
 	fr_process_end_command (comm->process);
 	fr_process_start (comm->process);
 }
 
 
-static void
-fr_command_rpm_set_mime_type (FrCommand  *comm,
-		 	      const char *mime_type)
+const char *rpm_mime_type[] = { "application/x-rpm", NULL };
+
+
+const char **
+fr_command_rpm_get_mime_types (FrCommand *comm)
 {
-	FR_COMMAND_CLASS (parent_class)->set_mime_type (comm, mime_type);
-	
-	comm->capabilities |= FR_COMMAND_CAP_ARCHIVE_MANY_FILES;
-	if (is_program_in_path ("rpm2cpio")) 
-		comm->capabilities |= FR_COMMAND_CAP_READ;
+	return rpm_mime_type;
 }
 
 
-static void 
+FrCommandCap
+fr_command_rpm_get_capabilities (FrCommand  *comm,
+			         const char *mime_type,
+				 gboolean    check_command)
+{
+	FrCommandCap capabilities;
+
+	capabilities = FR_COMMAND_CAN_ARCHIVE_MANY_FILES;
+	if (is_program_available ("cpio", check_command))
+		capabilities |= FR_COMMAND_CAN_READ;
+
+	return capabilities;
+}
+
+
+static const char *
+fr_command_rpm_get_packages (FrCommand  *comm,
+			     const char *mime_type)
+{
+	return PACKAGES ("cpio,rpm");
+}
+
+
+static void
 fr_command_rpm_class_init (FrCommandRpmClass *class)
 {
         GObjectClass   *gobject_class = G_OBJECT_CLASS (class);
@@ -215,16 +259,17 @@ fr_command_rpm_class_init (FrCommandRpmClass *class)
 
 	gobject_class->finalize = fr_command_rpm_finalize;
 
-        afc->list           = fr_command_rpm_list;
-	afc->extract        = fr_command_rpm_extract;
-	afc->set_mime_type  = fr_command_rpm_set_mime_type;
+        afc->list             = fr_command_rpm_list;
+	afc->extract          = fr_command_rpm_extract;
+	afc->get_mime_types   = fr_command_rpm_get_mime_types;
+	afc->get_capabilities = fr_command_rpm_get_capabilities;
+	afc->get_packages     = fr_command_rpm_get_packages;
 }
 
- 
-static void 
+
+static void
 fr_command_rpm_init (FrCommand *comm)
 {
-	comm->propCanModify                = FALSE;
 	comm->propAddCanUpdate             = FALSE;
 	comm->propAddCanReplace            = FALSE;
 	comm->propExtractCanAvoidOverwrite = FALSE;
@@ -235,7 +280,7 @@ fr_command_rpm_init (FrCommand *comm)
 }
 
 
-static void 
+static void
 fr_command_rpm_finalize (GObject *object)
 {
         g_return_if_fail (object != NULL);
